'use strict';

var fs = require('fs');
var net = require('net');

var ExpectedError = require('../modules/errors.js').ExpectedError;
var app = require('../modules/app.js');
var storage = require('../modules/storage.js').default;
var system = require('../modules/system.js');

var config = require('../config.json');
var packageData = require('../../package.json');

var TMP_FIRMWARE_PATH = '/tmp/ruffos.bin';
var spawn = require('child_process').spawn;

exports.ping = function (options, callback) {
    callback();
};

exports.preflight = function (options, callback) {
    callback(undefined, {
        features: {
            bytecode: true,
            commandWithStream: true
        },
        version: packageData.version,
        minRapVersion: config.minRapVersion,
        tokenSetup: !!storage.get('auth-token')
    });
};

exports.setToken = function (options, callback) {
    storage.set('auth-token', options.newToken);
    callback();
};

exports.verifyToken = function (options, callback) {
    // Verification would actually be done by routing logic.
    callback();
};

exports.rename = function (options, callback) {
    var name = options.name;
    storage.set('device-name', name);
    callback();
};

exports.setServerSettings = function (options, callback) {
    var httpPort = options.httpPort;

    if (httpPort) {
        storage.set('http-port', httpPort);
    }

    callback();
};

exports.upgrade = function (options, callback) {
    var remote = options.remote;
    var stream = options.stream;
    var checksum = options.checksum;

    if (stream) {
        console.log('Start downloading firmware...');
    } else {
        console.log('Start downloading firmware from ' + remote.ip + ':' + options.port + '...');

        stream = net.connect(options.port, remote.ip, function () {
            console.log('Connected to rap server.');
        });
    }

    var downloaded = 0;
    var hasError = false;

    fs.writeFileSync(TMP_FIRMWARE_PATH, '');

    stream.on('data', function (data) {
        downloaded += data.length;

        console.log('Downloaded ' + Math.floor(downloaded / 1024) + ' kB.');

        fs.appendFileSync(TMP_FIRMWARE_PATH, data);
        Duktape.gc();
    });

    stream.on('end', function () {
        console.log('Upgrade downloading ended.');

        if (hasError) {
            return;
        }

        validate(TMP_FIRMWARE_PATH, checksum);
    });

    stream.on('error', function (error) {
        hasError = true;
        callback(new ExpectedError('Failed to download firmware', 2, error));
    });

    function validate(binPath, checksum) {
        if (!checksum) {
            console.log('No checksum provided, skipping validation...');
            prepareFirmwareWriting(binPath);
            return;
        }

        try {
            console.log('Validating checksum (' + checksum + ')...');

            var cp = spawn('md5sum', ['-c', '-s'], {
                cwd: '/',
                stdio: ['pipe', 'ignore', 'ignore']
            });

            cp.stdin.write(checksum + '  ' /* double spaces */ + binPath);
            cp.stdin.end();

            cp.on('exit', function (code) {
                if (code === 0) {
                    console.log('Firmware checksum okay.');
                    prepareFirmwareWriting(binPath);
                } else {
                    callback(new ExpectedError('Corrupted firmware file', 2));
                }
            });
        } catch (error) {
            callback(new ExpectedError('Failed to write firmware', 2, error));
        }
    }

    function prepareFirmwareWriting(binPath) {
        var cp = spawn('/etc/init.d/ruffwifi', ['stop'], {
            cwd: '/',
            stdio: ['ignore', 'ignore', 'ignore']
        });

        cp.on('exit', function (code) {
            if (code === 0) {
                writeFirmware(binPath);
            } else {
                callback(new ExpectedError('Failed to stop ruffwifi', 2));
            }
        });
    }

    function writeFirmware(binPath) {
        try {
            console.log('Writing firmware...');

            var cp = spawn('mtd', ['write', binPath, 'firmware'], {
                cwd: '/',
                stdio: ['ignore', 'ignore', 'ignore']
            });

            cp.on('exit', function (code) {
                if (code === 0) {
                    console.log('Rebooting...');

                    spawn('reboot', []);
                    callback();
                } else {
                    callback(new ExpectedError('Failed to write firmware', 2));
                }
            });
        } catch (error) {
            callback(new ExpectedError('Failed to write firmware', 2, error));
        }
    }
};

exports.reboot = function (options, callback) {
    try {
        spawn('reboot', []);
        callback();
    } catch (error) {
        console.error(error);
        callback(new ExpectedError('Failed reboot device', 2));
    }
};

exports.reset = function (options, callback) {
    try {
        var handle = spawn('jffs2reset', ['-y'], {
            cwd: '/',
            stdio: ['ignore', 'ignore', 'ignore']
        });

        handle.on('exit', function (code) {
            if (code === 0) {
                spawn('reboot', []);
                callback();
            } else {
                callback(new ExpectedError('Failed to reboot device', 2));
            }
        });
    } catch (error) {
        callback(new ExpectedError('Failed to reset device', 2, error));
    }
};

exports.setWifi = function (options, callback) {
    var ssid = options.ssid;
    var password = options.password;

    try {
        spawn('/usr/sbin/enable_wifi_sta.sh', [ssid, password], {
            cwd: '/',
            stdio: ['ignore', 'ignore', 'ignore']
        });
        callback();
    } catch (error) {
        callback(new ExpectedError('Failed to configure Wifi settings', 2, error));
    }
};

exports.setAP = function (options, callback) {
    var password = options.password;

    try {
        spawn('/usr/sbin/set_ap_passwd.sh', [password], {
            cwd: '/',
            stdio: ['ignore', 'ignore', 'ignore']
        });
        callback();
    } catch (error) {
        callback(new ExpectedError('Failed to configure AP password', 2, error));
    }
};

exports.getNetStatus = function (options, callback) {
    var staStatus = {};
    var ethStatus = {};
    var netStatus = {};
    var status;

    var staStatusFilePath = '/etc/netmanager/sta_status';
    var ethStatusFilePath = '/etc/netmanager/eth_status';

    if (fs.existsSync(staStatusFilePath)) {
        status = fs.readFileSync(staStatusFilePath, 'utf-8')
                    .split('\n')
                    .map(function (line) {
                        return line.substr(line.indexOf(':') + 1).trim();
                    });
        staStatus.connected = status[0] === 'true';
        staStatus.ssid = status[1];
        staStatus.ip = status[2];
    } else {
        callback(new ExpectedError('Failed to get station status', 2));
    }

    if (fs.existsSync(ethStatusFilePath)) {
        status = fs.readFileSync(ethStatusFilePath, 'utf-8')
                    .split('\n')
                    .map(function (line) {
                        return line.substr(line.indexOf(':') + 1).trim();
                    });
        ethStatus.connected = status[0] === 'true';
        ethStatus.ip = status[1];
    } else {
        callback(new ExpectedError('Failed to get ethernet status', 2));
    }

    netStatus.sta = staStatus;
    netStatus.eth = ethStatus;

    callback(undefined, netStatus);
};

exports.getInfo = function (options, callback) {
    var builders = [
        function () {
            this.deviceName = storage.get('device-name');
        },
        function () {
            this.ruffVersion = fs
                .readFileSync('/etc/ruff_version', 'utf-8')
                .replace(/^RUFF_VERSION=/, '')
                .trim();
        },
        function () {
            var fd = uv.fs_open('/proc/version', 'rn', parseInt('666', 8));

            var text = uv
                .fs_read(fd, 1024, 0)
                .toString()
                .trim();

            uv.fs_close(fd);

            this.systemVersion = text;
        },
        function () {
            var timestamp = Math.floor(uv.hrtime() / 1e6);

            this.ruffdUptime = (timestamp - global.RUFFD_START_TIME) / 1000 || undefined;
            this.appUptime = (timestamp - app.startTime) / 1000 || undefined;

            this.systemUptime = system.getUptime();
        },
        function () {
            var macAddress = process.ruff.mac;
            this.macAddress = macAddress && macAddress.replace(/.{2}(?!$)/g, '$&:');
            this.serialNumber = process.ruff.sn;
        },
        function () {
            this.appStatus = app.status;
        },
        function () {
            if (this.status === 'not-deployed') {
                return;
            }

            var json = fs.readFileSync(app.packageFilePath, 'utf-8');
            var data = JSON.parse(json);

            this.appName = data.name;
            this.appVersion = data.version;
        }
    ];

    var info = {};

    builders.forEach(function (builder) {
        try {
            builder.call(info);
        } catch (error) {
            console.error(error);
        }
    });

    callback(undefined, info);
};
