'use strict';

var EventEmitter = require('events');
var util = require('util');

var MAGIC = 'SERIAL';
var POSTAMBLE = new Buffer('!');

var State = {
    magic: 0,
    sequence: 1,
    length: 2,
    body: 3
};

function Parser() {
    EventEmitter.call(this);

    this._state = State.magic;
    this._magicError = false;
    this._pending = new Buffer(0);
}

util.inherits(Parser, EventEmitter);

Object.defineProperty(Parser.prototype, 'pending', {
    get: function () {
        return this._pending.length > 0 || this._state !== State.magic;
    }
});

Parser.prototype.append = function (buffer) {
    this._pending = Buffer.concat([this._pending, buffer]);
    this._parse();
};

Parser.prototype._parse = function () {
    while (this._pending.length) {
        var parsed;

        switch (this._state) {
            case State.magic:
                parsed = this._parseMagic();
                break;
            case State.sequence:
                parsed = this._parseSequence();
                break;
            case State.length:
                parsed = this._parseLength();
                break;
            case State.body:
                parsed = this._parseBody();
                break;
            case State.checksum:
                parsed = this._parseChecksum();
                break;
        }

        if (!parsed) {
            break;
        }
    }
};

Parser.prototype._consume = function (length) {
    this._pending = this._pending.slice(length);
};

Parser.prototype._parseMagic = function () {
    var magic = this._pending.toString('binary', 0, MAGIC.length);

    for (var i = 0; i < magic.length; i++) {
        // Magic number should fail as soon as possible.
        if (magic[i] !== MAGIC[i]) {
            // Consume pending buffers.
            this._consume(i + 1);

            // Avoid too many errors.
            if (!this._magicError) {
                this._magicError = true;
                // this.emit('error', new Error('Invalid magic number'));
            }

            return true;
        }
    }

    if (magic.length === MAGIC.length) {
        this._consume(magic.length);
        this._state = State.sequence;
        this._magicError = false;

        return true;
    } else {
        // Insufficient magic length.
        return false;
    }
};

Parser.prototype._parseSequence = function () {
    if (this._pending.length < 2) {
        return false;
    }

    this._currentSequence = this._pending.readUInt16BE(0);
    this._currentPacket = this._pending.slice(0, 2);

    this._consume(2);

    this._state = State.length;

    return true;
};

Parser.prototype._parseLength = function () {
    if (this._pending.length < 2) {
        return false;
    }

    this._length = this._pending.readUInt16BE(0);
    this._currentPacket = Buffer.concat([
        this._currentPacket,
        this._pending.slice(0, 2)
    ]);

    this._consume(2);
    this._state = State.body;

    return true;
};

Parser.prototype._parseBody = function () {
    var length = this._length;

    if (this._pending.length < (length + 1)) {
        return false;
    }

    var body = this._pending.slice(0, length);
    var checksum = this._pending.readUInt8(length, length + 1);

    this._currentPacket = Buffer.concat([
        this._currentPacket,
        body
    ]);

    // console.log('length is ', this._length, this._currentPacket.length);
    // console.log('decoded body is ', this._currentPacket.toString('hex'), this._currentPacket.toString());
    // console.log('checksum is', this._pending.slice(length, length + 1), checksum, calChecksum(this._currentPacket));
    if (checksum === calChecksum(this._currentPacket)) {
        this.emit('data', body, this._currentSequence);
    } else {
        this.emit('error', new Error('checksum error'), this._currentSequence);
    }

    this._consume(length + 1);
    this._state = State.magic;

    return true;
};

exports.Parser = Parser;

function encodeMagic() {
    return new Buffer(MAGIC);
}

function encodeLength(length) {
    var buffer = new Buffer(2);
    buffer.writeUInt16BE(length);
    return buffer;
}

function encodeSequence(sequence) {
    var buffer = new Buffer(2);
    buffer.writeUInt16BE(sequence);
    return buffer;
}

function encodeBody(data) {
    if (typeof data === 'string') {
        return new Buffer(data);
    }
    return data;
}

function calChecksum(data) {
    var checksum = 0;
    for (var i = 0; i < data.length; i++) {
        checksum = (checksum + data[i]) & 0xFF;
    }
    return checksum;
}

function build(sequence, data) {
    var magicNumberBuffer = encodeMagic();
    var sequenceBuffer = encodeSequence(sequence);
    var bodyBuffer = encodeBody(data);
    var bodyLengthBuffer = encodeLength(bodyBuffer.length);
    var packet = Buffer.concat([
        sequenceBuffer,
        bodyLengthBuffer,
        bodyBuffer
    ]);
    var checksum = calChecksum(packet);
    var totalPacket = Buffer.concat([
        magicNumberBuffer,
        packet,
        new Buffer([checksum]),
        POSTAMBLE
    ]);

    return totalPacket;
}

exports.build = build;
